#!/usr/local/sbin/charm-env python3
"""Charm actions."""

import os.path
import sys
import traceback
from subprocess import CalledProcessError, check_output

from charmhelpers.core import hookenv


def activate(params):
    """Execure activate action."""
    config = hookenv.config()
    livepatch_key = config.get("livepatch_key")

    if livepatch_key:
        # disable prior to enabling to work around LP#1628823
        cmd = ["/snap/bin/canonical-livepatch", "disable"]
        try:
            check_output(cmd, universal_newlines=True)
        except CalledProcessError as e:
            hookenv.log("Unable to deactivate: {}".format(str(e)))
            # but let's soldier on...
        cmd = [
            "/snap/bin/canonical-livepatch",
            "enable",
            "{}".format(livepatch_key.strip()),
        ]
        try:
            check_output(cmd, universal_newlines=True)
        except CalledProcessError as e:
            hookenv.action_fail("Unable to activate: {}".format(str(e)))
            return
        else:
            hookenv.action_set(dict(result="Activated"))

    else:
        hookenv.action_fail("Unable to activate as no key has been set")


def deactivate(params):
    """Execute deactivate action."""
    cmd = ["/snap/bin/canonical-livepatch", "disable"]
    try:
        check_output(cmd, universal_newlines=True)
    except CalledProcessError as e:
        hookenv.action_fail("Unable to deactivate: {}".format(str(e)))
        return
    hookenv.action_set(dict(result="Deactivated"))


def refresh(params):
    """Execute refresh action."""
    cmd = ["/snap/bin/canonical-livepatch", "refresh"]
    try:
        check_output(cmd, universal_newlines=True)
    except CalledProcessError as e:
        hookenv.action_fail("Unable to refresh: {}".format(str(e)))
        return
    hookenv.action_set(dict(result="Refreshed"))


def main(argv):
    """Select the appropriate action."""
    action = os.path.basename(argv[0])
    params = hookenv.action_get()
    try:
        if action == "activate":
            activate(params)
        elif action == "deactivate":
            deactivate(params)
        elif action == "refresh":
            refresh(params)
        else:
            hookenv.action_fail("Action {} not implemented".format(action))
    except Exception:
        hookenv.action_fail("Unhandled exception")
        tb = traceback.format_exc()
        hookenv.action_set(dict(traceback=tb))
        hookenv.log("Unhandled exception in action {}".format(action))
        print(tb)


if __name__ == "__main__":
    main(sys.argv)
