#!/usr/bin/env python3
"""Check canonical-livepatch and alert."""

# Copyright (C) 2016 Canonical Ltd.

import os
from subprocess import call, check_output

import nagios_plugin3

from yaml import safe_load

supported_archs = ["x86_64"]


def check_snap_installed():
    """Confirm the snap is installed, raise an error if not."""
    cmd = ["snap", "list", "canonical-livepatch"]
    try:
        check_output(cmd).decode("UTF-8")
    except Exception:
        raise nagios_plugin3.CriticalError("canonical-livepatch snap is not installed")


def load_status():
    """Load the cached status from disk, return it as a string."""
    livepatch_output_path = "/var/lib/nagios/canonical-livepatch-status.txt"

    with open(livepatch_output_path, "r") as canonical_livepatch_log:
        livepatch_status = canonical_livepatch_log.read()

    return livepatch_status.strip()


def check_serious_errors():
    """Check serious error cases, the output will not be YAML."""
    livepatch_status = load_status()

    # if it's empty, something's definitely wrong
    if livepatch_status == "":
        raise nagios_plugin3.CriticalError("No output from canonical-livepatch status")

    # in some cases, it's obviously not valid YAML
    try:
        livepatch_yaml = safe_load(livepatch_status)
    except Exception:
        raise nagios_plugin3.CriticalError(livepatch_status.replace("\n", " "))

    # in other cases, it's less obvious until we try to parse
    try:
        livepatch_yaml["status"]
    except TypeError:
        raise nagios_plugin3.CriticalError(livepatch_status.replace("\n", " "))


def active_kernel_version():
    """Return the active kernel version, from livepatch's perspective."""
    livepatch_status = load_status()
    status_yaml = safe_load(livepatch_status)
    for kernel in status_yaml.get("status"):
        if kernel.get("running") is True:
            return kernel.get("kernel")


def check_status():
    """Check the cached status, raise an error if we find any issues."""
    livepatch_status = load_status()
    errors = []

    status_yaml = safe_load(livepatch_status)

    for kernel in status_yaml.get("status"):
        if kernel.get("running") is True:
            check_state = kernel.get("livepatch").get("checkState")
            patch_state = kernel.get("livepatch").get("patchState")

    check_state_error = check_check_state(check_state)
    patch_state_error = check_patch_state(patch_state)

    if check_state_error:
        errors.append(check_state_error)
    if patch_state_error:
        errors.append(patch_state_error)

    if errors:
        raise nagios_plugin3.CriticalError(" ".join(errors))


def check_check_state(check_state):
    """Check for issues with checkState, including unexpected output."""
    if check_state in ["checked", "needs-check"]:
        pass
    elif check_state == "check-failed":
        return "Livepatch failed to check the remote service for patches."
    else:
        return "Unknown check state: {}".format(check_state)


def check_patch_state(patch_state):
    """Check for issues with patchState, including unexpected output."""
    if patch_state in ["applied", "applying", "nothing-to-apply"]:
        pass
    elif patch_state == "unapplied":
        return "Livepatch has not applied the downloaded patches."
    elif patch_state == "applied-with-bug":
        return "Livepatch has detected a kernel bug while applying patches."
    elif patch_state == "apply-failed":
        return "Livepatch failed to apply patches."
    elif patch_state == "kernel-upgrade-required":
        return "A kernel upgrade (and reboot) is required."
    else:
        return "Unknown patch state: {}".format(patch_state)


def lsb_release():
    """Return /etc/lsb-release in a dict."""
    d = {}
    with open("/etc/lsb-release", "r") as lsb:
        for line in lsb:
            k, v = line.split("=")
            d[k.strip()] = v.strip()
    return d


def init_is_systemd():
    """Return True if the host system uses systemd, False otherwise."""
    if lsb_release()["DISTRIB_CODENAME"] == "trusty":
        return False
    return os.path.isdir("/run/systemd/system")


def is_container():
    """Determine whether unit is running in a container.

    @return: boolean indicating if unit is in a container
    """
    if init_is_systemd():
        # Detect using systemd-detect-virt
        return call(["systemd-detect-virt", "--container", "--quiet"]) == 0
    else:
        # Detect using upstart container file marker
        return os.path.exists("/run/container_type")


def main():
    """Run all the checks."""
    arch = os.uname()[4]
    if arch not in supported_archs:
        raise nagios_plugin3.CriticalError(
            "canonical-livepatch not supported on this architecture ({}).".format(arch)
        )
    elif is_container():
        print("canonical-livepatch not needed in OS containers.")
    else:
        nagios_plugin3.try_check(check_snap_installed)
        nagios_plugin3.try_check(check_serious_errors)
        nagios_plugin3.try_check(check_status)
        kernel_version = active_kernel_version()
        print("OK - canonical-livepatch is active on kernel {}".format(kernel_version))


if __name__ == "__main__":
    main()
