"""Charm hooks."""

from distutils.version import LooseVersion
from os import environ, mkdir, path, uname
from subprocess import CalledProcessError, check_call, check_output
from time import sleep

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv
from charmhelpers.core.host import is_container, write_file

from charms.layer import snap
from charms.reactive import clear_flag, hook, set_flag, when, when_not
from charms.reactive.flags import register_trigger

from yaml import safe_load


def file_to_units(local_path, unit_path):
    """Copy a file from the charm onto our unit(s)."""
    if local_path[-3:] == ".py":
        perms = 0o755
    else:
        perms = 0o644

    fh = open(local_path, "r")
    write_file(
        path=unit_path, content=fh.read().encode(), owner="root", perms=perms,
    )


def wait_for_path(file_path, timeout=30):
    """Wait until path exists or timeout."""
    hookenv.log("Waiting for up to {}s for {}".format(timeout, file_path))
    seconds_waited = 0
    interval = 1
    while not (path.exists(file_path)):
        hookenv.log("{} does not exist - waiting {}s".format(file_path, interval))
        sleep(interval)
        seconds_waited += interval
        if seconds_waited >= timeout:
            hookenv.log("Giving up waiting for {}".format(file_path), hookenv.ERROR)
            return
    return


def wait_for_livepatch():
    """Wait until canonical-livepatch socket is created."""
    wait_for_path("/var/snap/canonical-livepatch/current/livepatchd-priv.sock")


def unit_update(status=None, message=None):
    """Update status message."""
    if status and message:
        hookenv.status_set(status, message)
    else:
        patch_details = get_patch_details()
        hookenv.status_set(
            "active", "Running kernel {}, patchState: {}".format(*patch_details)
        )


def get_livepatch_status():
    """Get livepatch status."""
    cmd = ["/snap/bin/canonical-livepatch", "status", "--format", "yaml"]
    livepatch_status = ""
    try:
        livepatch_status = check_output(cmd, universal_newlines=True)
    except FileNotFoundError:
        # If the snap's not installed yet, don't crash out
        pass
    except CalledProcessError:
        # If the service hasn't been enabled yet,
        # we'll get a process error - this is fine
        pass
    return livepatch_status


def write_status_to_disk():
    """Write livepatch status to disk."""
    # In a race with nrpe, create /var/lib/nagios if nrpe didn't do it yet
    if not path.exists("/var/lib/nagios"):
        mkdir("/var/lib/nagios")

    current_status = get_livepatch_status()
    perms = 0o644
    write_file(
        path="/var/lib/nagios/canonical-livepatch-status.txt",
        content=current_status,
        owner="root",
        perms=perms,
    )


def get_patch_details():
    """Get patch details.

    :return: a tuple containing kernel version and patch state.
    """
    kernel = "unknown"
    patch_state = "unknown"
    raw_status = get_livepatch_status()

    # status will usually pass YAML (but not always!)
    try:
        status = safe_load(raw_status)
    except Exception:
        hookenv.log("Unable to parse status yaml", hookenv.ERROR)
        return kernel, raw_status

    # even if we got YAML, be paranoid
    try:
        for k in status["status"]:
            if k["running"] is True:
                kernel = k["kernel"]
                patch_state = k["livepatch"]["patchState"]
    except Exception:
        hookenv.log("Unable to find patching details in status yaml", hookenv.ERROR)
        return kernel, patch_state

    return kernel, patch_state


def configure_proxies(proxy=None):
    """Configure proxies."""
    for proto in ("http-proxy", "https-proxy"):
        if proxy is None:
            proxy = ""
        cmd = ["/snap/bin/canonical-livepatch", "config", "{}={}".format(proto, proxy)]
        hookenv.log("Configuring {} as {}".format(proto, proxy))
        try:
            check_call(cmd, universal_newlines=True)
        except CalledProcessError:
            hookenv.log("Failed to set proxy", hookenv.ERROR)


def activate_livepatch():
    """Activate livepatch."""
    unit_update("maintenance", "Updating API key")
    config = hookenv.config()

    livepatch_key = config.get("livepatch_key")
    if livepatch_key:
        # disable prior to enabling to work around LP#1628823
        cmd = ["/snap/bin/canonical-livepatch", "disable"]
        hookenv.log(
            "Disabling canonical-livepatch ahead "
            "of key change to work around LP#1628823"
        )
        try:
            check_output(cmd, universal_newlines=True)
        except CalledProcessError as e:
            hookenv.log("Unable to deactivate: {}".format(str(e)), hookenv.ERROR)

        cmd = [
            "/snap/bin/canonical-livepatch",
            "enable",
            "{}".format(livepatch_key.strip()),
        ]
        hookenv.log("Activating canonical-livepatch with your key")
        try:
            check_output(cmd, universal_newlines=True)
        except CalledProcessError as e:
            hookenv.log("Unable to activate: {}".format(str(e)), hookenv.ERROR)
            clear_flag("canonical-livepatch.active")
            unit_update("blocked", "Activation failed")
        else:
            set_flag("canonical-livepatch.active")
            unit_update()
    else:
        hookenv.log(
            "Unable to activate canonical-livepatch as no key has been set",
            hookenv.ERROR,
        )
        cmd = ["/snap/bin/canonical-livepatch", "disable"]
        try:
            check_output(cmd, universal_newlines=True)
        except CalledProcessError as e:
            hookenv.log("Unable to deactivate: {}".format(str(e)), hookenv.ERROR)
        clear_flag("canonical-livepatch.active")
        unit_update("blocked", "Service disabled, please set livepatch_key to activate")


register_trigger(
    when="config.changed.livepatch_proxy", clear_flag="livepatch-proxy.configured"
)


@when_not("canonical-livepatch.supported")
def livepatch_supported():
    """Check if livepatch is supported and set flag."""
    arch = uname()[4]
    supported_archs = ["x86_64"]

    if arch not in supported_archs:
        hookenv.log("Livepatch does not currently support {} architecture".format(arch))
        unit_update(
            "blocked", "Architecture {} is not supported by livepatch".format(arch)
        )
    elif is_container() and not hookenv.config()["enable_container"]:
        hookenv.log("OS container detected, livepatch is not needed")
        unit_update("blocked", "Livepatch is not needed in OS containers")
    else:
        set_flag("canonical-livepatch.supported")


@when("canonical-livepatch.supported")
@when_not("snap.installed.canonical-livepatch")
def install_livepatch():
    """Install livepatch."""
    config = hookenv.config()
    snap_channel = config.get("snap_channel")
    snap.install("canonical-livepatch", **{"channel": snap_channel})


@when("snap.installed.canonical-livepatch")
@when_not("livepatch-proxy.configured")
def proxy_settings():
    """Get proxy settings and configure it."""
    # Configure proxies early, if required - LP#1761661

    preferred_proxy = None  # default to None

    for key, value in environ.items():
        # use environment's https_proxy or http_proxy if either are set
        if key == "https_proxy":
            preferred_proxy = value
        elif key == "http_proxy":
            preferred_proxy = value

    # but if we have a charm configured proxy, prefer that
    config = hookenv.config()
    charm_config_proxy = config.get("livepatch_proxy")
    if charm_config_proxy:
        preferred_proxy = charm_config_proxy

    # whatever we prefer (or None), configure it!
    configure_proxies(preferred_proxy)

    set_flag("livepatch-proxy.configured")


@when("snap.installed.canonical-livepatch")
@when("livepatch-proxy.configured")
@when_not("canonical-livepatch.connected")
def canonical_livepatch_connect():
    """Check if livepatch socket is created and update unit."""
    # So if we've just installed snapd on a trusty system, we will not be on
    # the HWE kernel yet and unfortunately need to reboot first!
    current = LooseVersion(uname()[2])
    required = LooseVersion("4.4")
    uptrack_path = "/usr/sbin/uptrack-upgrade"
    if path.exists(uptrack_path):
        hookenv.log("Ksplice/Uptrack detected, please remove it and reboot")
        unit_update("blocked", "Remove ksplice and then reboot")
    elif current < required:
        hookenv.log("Reboot required, kernel {} is too old".format(current))
        unit_update("blocked", "A reboot is required")
    else:
        unit_update("maintenance", "Connecting to the livepatch service")
        # Make sure the service is ready for us
        wait_for_livepatch()
        set_flag("canonical-livepatch.connected")
        unit_update("blocked", "Service disabled, please set livepatch_key to activate")


@when("canonical-livepatch.connected")
@when_not("config.changed.livepatch_key", "canonical-livepatch.active")
def init_key():
    """Activate livepatch."""
    # If deployed under Trusty before rebooting into the HWE kernel
    # the config-changed hook won't fire post-reboot as the state
    # isn't tracked, but we didn't initialise yet! So, handle it here
    activate_livepatch()


@when("canonical-livepatch.connected", "config.changed.livepatch_key")
def update_key():
    """Activate livepatch and write status to disk."""
    # handle regular config-changed hooks for the livepatch key
    activate_livepatch()
    write_status_to_disk()


@when("snap.installed.canonical-livepatch", "config.changed.snap_channel")
def change_channel():
    """Update snap channel."""
    config = hookenv.config()
    snap_channel = config.get("snap_channel")
    # refresh to the given channel
    snap.refresh("canonical-livepatch", **{"channel": snap_channel})


# Set up Nagios checks when the nrpe-external-master subordinate is related
@when("nrpe-external-master.available")
@when_not("canonical-livepatch.nagios-setup.complete")
@when("canonical-livepatch.supported")
def configure_nagios(nagios):
    """Configure nagios checks."""
    if hookenv.hook_name() == "update-status":
        return

    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=False)

    # install nagios support files
    file_to_units(
        "files/check_canonical-livepatch.cron", "/etc/cron.d/check_canonical-livepatch",
    )
    file_to_units(
        "files/check_canonical-livepatch.py",
        "/usr/lib/nagios/plugins/check_canonical-livepatch.py",
    )

    # write current status to disk to satisfy the nagios check
    write_status_to_disk()

    # remove check from previous release with poorly formed name
    nrpe_setup.remove_check(shortname="check_canonical-livepatch")

    # use charmhelpers to create the check
    nrpe_setup.add_check(
        "canonical-livepatch",
        "Verify canonical-livepatch is working",
        "/usr/lib/nagios/plugins/check_canonical-livepatch.py",
    )

    nrpe_setup.write()

    # Remove obsolete state from older charm, if present
    clear_flag("canonical-livepatch.nagios-configured")
    # Set our preferred state
    set_flag("canonical-livepatch.nagios-setup.complete")


@when("snap.installed.canonical-livepatch", "canonical-livepatch.active")
def update_kernel_version():
    """Update kernel version."""
    unit_update()


# This is triggered on any config-changed, and after an upgrade-charm - you
# don't get the latter with @when('config.changed')
@hook("config-changed")
def set_nrpe_flag():
    """Run config changed."""
    clear_flag("canonical-livepatch.nagios-setup.complete")
