"""Zaza functional tests."""
import logging
import unittest

import yaml

import zaza.model as model


class TestBase(unittest.TestCase):
    """Base Class for charm functional tests."""

    @classmethod
    def setUpClass(cls):
        """Run setup for tests."""
        cls.model_name = model.get_juju_model()
        cls.application_name = "canonical-livepatch"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        cls.units = model.get_units(cls.application_name, model_name=cls.model_name)

        cls.lead_unit = cls.units[0]

        cls.canonical_livepatch_ip = model.get_app_ips(cls.application_name)[0]


class TestCanonicalLivepatch(TestBase):
    """Class for charm functional tests."""

    def test_01_install(self):
        """Verify that the snap is installed."""
        if self.lead_unit.workload_status_message == (
            "Livepatch is not needed in OS containers"
        ):
            model.set_application_config(
                self.application_name, {"enable_container": "true"}
            )
        model.block_until_all_units_idle()
        cmd = "stat /snap/bin/canonical-livepatch"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "0")

    def test_02_status(self):
        """Verify canonical-livepatch status."""
        # run a status check - we expect this to return 1 due to no access key
        cmd = "sudo canonical-livepatch status"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "1")

    def test_03_nagios_init(self):
        """Verify nagios plugins are rendered."""
        for path in [
            "/etc/cron.d/check_canonical-livepatch",
            "/usr/lib/nagios/plugins/check_canonical-livepatch.py",
            "/var/lib/nagios/canonical-livepatch-status.txt",
        ]:
            logging.info("Checking that file '{}' exists".format(path))
            cmd = "stat {}".format(path)
            result = model.run_on_unit(self.lead_unit_name, cmd)
            code = result.get("Code")
            self.assertEqual(code, "0")

    def test_04_nagios_context_update(self):
        """Update nagios context and verify."""
        test_context_name = "new_context"
        model.set_application_config(
            self.application_name, {"nagios_context": test_context_name}
        )
        model.block_until_all_units_idle()

        cmd = (
            "grep {} /var/lib/nagios/export/"
            "service__*_check_canonical-livepatch.cfg".format(test_context_name)
        )

        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "0")

    def test_05_channel_update(self):
        """Update snap channel and verify."""
        # verify the current channel
        cmd = "snap info canonical-livepatch"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "0")
        output = result.get("Stdout")

        # confirm we're tracking 'stable'
        output_yaml = yaml.safe_load(output)
        channel = output_yaml["tracking"]
        self.assertIn("stable", channel)

        # change channel to 'beta'
        model.set_application_config(self.application_name, {"snap_channel": "beta"})
        model.block_until_all_units_idle()

        # verify the current channel
        cmd = "snap info canonical-livepatch"
        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "0")
        output = result.get("Stdout")

        # confirm we're tracking 'beta'
        output_yaml = yaml.safe_load(output)
        channel = output_yaml["tracking"]
        self.assertIn("beta", channel)

    def test_06_nagios_servicegroup_update(self):
        """Update nagios service group and verify."""
        test_servicegroup_name = "new_servicegroup"
        model.set_application_config(
            self.application_name, {"nagios_servicegroups": test_servicegroup_name}
        )
        model.block_until_all_units_idle()

        cmd = (
            "grep {} /var/lib/nagios/export/"
            "service__*_check_canonical-livepatch.cfg".format(test_servicegroup_name)
        )

        result = model.run_on_unit(self.lead_unit_name, cmd)
        code = result.get("Code")
        self.assertEqual(code, "0")
